import http.client
import json

SUPPORTED_LANGUAGES = ["go", "python"]


class LangtoolsDaemonClient:
    def __init__(self, host="localhost", port=61782):
        self.host = host
        self.port = port

    def validate_language(self, language: str):
        if language not in SUPPORTED_LANGUAGES:
            raise NotImplementedError(
                f"No analyzer registered for language: {language!r}"
            )

    def analyze(self, language: str, project_root: str):
        self.validate_language(language)
        conn = http.client.HTTPConnection(self.host, self.port, timeout=60)
        data = json.dumps({"language": language, "project_root": project_root})
        conn.request(
            "POST", "/", body=data, headers={"Content-Type": "application/json"}
        )
        resp = conn.getresponse()
        resp_data = resp.read()
        return json.loads(resp_data)



import logging
import sys


def setup_logging():
    logging.basicConfig(
        level=logging.DEBUG,
        stream=sys.stderr,
        format="%(asctime)s %(levelname)s [%(name)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )



import json
from typing import Dict, List


def parse_pyright_output(output: str) -> List[Dict]:
    """
    Parses the JSON output from Pyright, extracting the list of
    diagnostics from the 'generalDiagnostics' key.
    """
    if not output.strip():
        return []

    data = json.loads(output)
    return data.get("generalDiagnostics", [])


def parse_as_json_document(output: str) -> List[Dict]:
    """Parses a single JSON document (e.g., a list of issues)."""
    if not output.strip():
        return []
    return json.loads(output)



import logging
import os
import shutil
import subprocess
from typing import Any, Callable, Dict, List, Union

logger = logging.getLogger(__name__)


class ToolRunner:
    def __init__(self, cwd: str, bin_dir: Union[str, List[str], None] = None):
        self.cwd = cwd
        if bin_dir is None:
            self.bin_dirs = []
        elif isinstance(bin_dir, str):
            self.bin_dirs = [bin_dir]
        else:
            self.bin_dirs = list(bin_dir)

    def run(
        self, cmd: List[str], parser: Callable[[str], Any] = lambda x: x
    ) -> List[Dict]:
        try:
            env = os.environ.copy()
            if self.bin_dirs:
                env["PATH"] = os.pathsep.join(self.bin_dirs + [env.get("PATH", "")])
            # Resolve executable using PATH
            tool_path = shutil.which(cmd[0], path=env["PATH"])
            if tool_path:
                cmd[0] = tool_path
            logger.debug(
                f"Running: {' '.join(cmd)} (CWD={self.cwd}) (PATH={env['PATH']})"
            )
            proc = subprocess.run(
                cmd,
                cwd=self.cwd,
                capture_output=True,
                encoding="utf-8",
                check=False,
            )
            output = proc.stdout if proc.stdout else proc.stderr
            return parser(output)

        except FileNotFoundError:
            tool_name = cmd[0]
            logger.error(f"Error: The tool '{tool_name}' was not found.")
            return [
                {
                    "source": "daemon_error",
                    "file": tool_name,
                    "message": f"Analysis tool '{tool_name}' is not installed or not in PATH.",
                }
            ]
        except Exception as e:
            logger.error(
                f"Failed to execute or parse for command '{' '.join(cmd)}': {e}"
            )
            return [
                {
                    "source": "daemon_error",
                    "file": cmd[0],
                    "message": str(e),
                }
            ]



import json
import logging
import os
from http.server import BaseHTTPRequestHandler, HTTPServer

from langtools_mcp.langtools.strategies import LANGUAGE_STRATEGIES
from langtools_mcp.logger import setup_logging

HOST = "localhost"
PORT = 61782

setup_logging()
logger = logging.getLogger("langtools_daemon")


class LangtoolsDaemonHandler(BaseHTTPRequestHandler):
    def send_error_json(self, code, message):
        self.send_response(code)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps({"status": "fail", "error": message}).encode())

    def do_POST(self):
        try:
            length = int(self.headers.get("Content-Length", 0))
            data = self.rfile.read(length)
            req = json.loads(data)
            language = req.get("language")
            project_root = req.get("project_root")
            logger.info(
                f"POST request: language={language}, project_root={project_root}"
            )
            if not language or not project_root:
                self.send_error_json(400, "Missing language or project_root")
                return
            try:
                strategy_cls = LANGUAGE_STRATEGIES[language.lower()]
                strategy = strategy_cls(project_root)
            except KeyError:
                self.send_error_json(400, f"Unsupported language: {language}")
                return
            result = strategy.analyze()
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(result.model_dump_json().encode())
        except Exception as exc:
            logger.exception("Unhandled exception")
            self.send_error_json(500, str(exc))


def run():
    host = os.getenv("LANGTOOLSD_HOST", HOST)
    port = int(os.getenv("LANGTOOLSD_PORT", PORT))
    server_address = (host, port)
    httpd = HTTPServer(server_address, LangtoolsDaemonHandler)
    logger.info(f"Langtools Daemon started on {host}:{port}")
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        logger.info("Shutting down daemon...")
        httpd.server_close()


if __name__ == "__main__":
    run()



import os
from pathlib import Path
from typing import Literal

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

# XDG Base Directory paths with fallbacks per spec
XDG_CONFIG_HOME = Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config"))
XDG_DATA_HOME = Path(os.environ.get("XDG_DATA_HOME", Path.home() / ".local" / "share"))
XDG_CACHE_HOME = Path(os.environ.get("XDG_CACHE_HOME", Path.home() / ".cache"))
XDG_STATE_HOME = Path(
    os.environ.get("XDG_STATE_HOME", Path.home() / ".local" / "state")
)
APP_NAME = "langtools_mcp"
BIN_DIR = XDG_DATA_HOME / "langtools_mcp" / "bin"


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="LANGTOOLS_")
    BIN_DIR: str = Field(
        description="Binary Directory to install any necessary tools",
        default=str(BIN_DIR),
    )


class PythonToolSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="LANGTOOLS_PYTHON")
    TOOLS: list[Literal["ruff", "pyright"]] = Field(
        description="Python tools to use. To see supported tools refer to documentation",
        default=["ruff", "pyright"],
    )

    BIN_DIR: str = Field(
        description="Bin directory for python tools",
        default=str(BIN_DIR),
    )


class GoToolSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="LANGTOOLS_GO")
    GO_TOOLS: list[Literal["vet"]] = Field(
        description="Go tools to use on your project. To see supported tools refer to documentation",
        default=["vet"],
    )

    BIN_DIR: str = Field(
        description="Bin directory for go tools",
        default=str(BIN_DIR),
    )


if __name__ == "__main__":
    os.environ["LANGTOOLS_PYTHON_TOOLS"] = '["ruff"]'
    settings = Settings()
    print(settings)



from langtools_mcp.langtools.langtools_daemon_client import LangtoolsDaemonClient


def run_analysis_for_language(language: str, project_root: str) -> dict:
    client = LangtoolsDaemonClient()
    return client.analyze(language, project_root)



import logging
from typing import Literal

from mcp.server.fastmcp import FastMCP
from mcp.shared.exceptions import McpError
from mcp.types import INVALID_REQUEST, ErrorData
from pydantic import BaseModel

from langtools_mcp.langtools.analysis import run_analysis_for_language
from langtools_mcp.logger import setup_logging

setup_logging()
logger = logging.getLogger(__name__)

INSTRUCTIONS = """
currently supports the following languages:
    - python
    - golang
    - typescript/javascript
When passing a `project_root` you MUST pass a full absolute path to the root of the project you are analyzing.
For monorepos, be sure to pass the project within the repo you want analysis on. 
"""

mcp = FastMCP("MCP to allow llms to analyze their code", INSTRUCTIONS)


class AnalyzeFileParams(BaseModel):
    language: Literal["python", "go", "typescript", "javascript"]
    project_root: str


@mcp.tool(
    "AnalyzeCodebase",
    description="Run a codebase through analysis for a given language. ",
)
def analyze_codebase(params: AnalyzeFileParams):
    try:
        analysis_result = run_analysis_for_language(
            language=params.language, project_root=params.project_root
        )
    except ValueError as e:
        raise McpError(ErrorData(message=str(e), code=INVALID_REQUEST))
    except NotImplementedError as e:
        raise McpError(ErrorData(message=str(e), code=INVALID_REQUEST))
    return analysis_result



import os
from collections import deque
from pathlib import Path
from typing import Iterable, Optional


class NoRootFoundException(Exception): ...


def search_down_for_roots(
    root_path: str,
    targets: Iterable[str],
    max_depth: int = 5,
    ignore: Iterable[str] = (),
) -> list[str]:
    """
    Recursively search downward from root_path for directories containing any of the target files/dirs.
    Skips directories or files in `ignore`.
    Returns a list of matching directory paths.
    """

    root = Path(root_path).resolve()
    found = []
    queue = deque([(root, 0)])
    ignore_set = set(ignore)

    while queue:
        curr, depth = queue.popleft()
        if depth > max_depth:
            continue
        # Ignore current directory if in ignore list
        if curr.name in ignore_set:
            continue
        # Check for targets in current directory
        for target in targets:
            if (curr / target).exists():
                found.append(str(curr))
                break
        # Queue subdirs, skipping those in ignore set
        for child in curr.iterdir():
            if (
                child.is_dir()
                and child.name not in ignore_set
                and not child.name.startswith(".")
            ):
                queue.append((child, depth + 1))
    if len(found) == 0:
        raise NoRootFoundException(f"No root found for the targets provided: {targets}")
    return found


def search_down_for_root(
    root_path: str,
    targets: Iterable[str],
    max_depth: int = 5,
    ignore: Iterable[str] = (),
) -> str:
    out = search_down_for_roots(root_path, targets, max_depth, ignore)
    return out[0]


def find_ts_root(path: str):
    return search_down_for_root(
        path,
        {"tsconfig.json", "package.json", "jsconfig.json"},
        ignore={"node_modules"},
    )


def find_virtual_env(path: str, max_depth=4) -> str | None:
    """
    Recursively search downward from `path` (up to max_depth levels)
    for a virtual environment directory.
    Looks for typical names and activation files.
    """
    root = Path(path)
    candidate_names = {".venv", "venv", "env", ".env"}
    # Breadth-first search up to max_depth
    queue = [(root, 0)]
    while queue:
        curr, depth = queue.pop(0)
        if depth > max_depth:
            continue
        for child in curr.iterdir():
            if child.is_dir() and child.name in candidate_names:
                # Check for venv activation script (Unix or Windows)
                if (child / "bin" / "activate").exists() or (
                    child / "Scripts" / "activate.bat"
                ).exists():
                    return str(child)
            # Enqueue subdirectories to search further down
            if child.is_dir() and not child.name.startswith("."):
                queue.append((child, depth + 1))
    return None


def find_go_module_root(path: str) -> str:
    path = os.path.abspath(path)

    # If it's a file, get its containing directory
    if os.path.isfile(path):
        dir_path = os.path.dirname(path)
    else:
        dir_path = path

    while True:
        maybe_mod = os.path.join(dir_path, "go.mod")
        if os.path.isfile(maybe_mod):
            return dir_path
        parent = os.path.dirname(dir_path)
        if parent == dir_path:
            break  # Reached the filesystem root
        dir_path = parent

    # Fallback: return starting directory (normalized)
    return os.path.abspath(path)



import logging
import os
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Literal, Type

from pydantic import BaseModel

from langtools_mcp.langtools.parsers import (
    parse_as_json_document,
    parse_pyright_output,
)
from langtools_mcp.langtools.settings import GoToolSettings, PythonToolSettings
from langtools_mcp.langtools.tool_runner import ToolRunner
from langtools_mcp.langtools.utils import (
    NoRootFoundException,
    find_go_module_root,
    find_ts_root,
    find_virtual_env,
)

logger = logging.getLogger(__name__)


class ToolSetupError(Exception): ...


class UnsupportedLanguageException(Exception): ...


class Diagnostic(BaseModel):
    status: Literal["ok", "failure"]
    source: str
    output: Any


class AnalysisResponse(BaseModel):
    status: Literal["ok", "fail"]
    diagnostics: list[Diagnostic]


class LanguageStrategy(ABC):
    def __init__(self, project_root: str):
        self.project_root = project_root
        self.available_tools: Dict[str, Callable[[], Diagnostic]] = {}

    def analyze(self) -> AnalysisResponse:
        logger.debug(f"Analyzing entire Python project at root: {self.project_root}")
        normalized_results = []
        for tool in self.configured_tools:
            normalized_results.append(self.call_tool_safely(tool))
        return AnalysisResponse(status="ok", diagnostics=normalized_results)

    def call_tool_safely(self, tool_name: str):
        try:
            return self.available_tools[tool_name]()
        except KeyError:
            logger.exception(f"Unable to find tool for {tool_name}")
            return Diagnostic(
                status="ok",
                source=tool_name,
                output="Unable to find tool. Analysis was skipped for this tool",
            )

    @property
    @abstractmethod
    def configured_tools(self) -> List[Any]: ...


class TypescriptStrategy(LanguageStrategy):
    def analyze(self) -> Any:
        try:
            root = find_ts_root(self.project_root)
        except NoRootFoundException:
            return {
                "status": "fail",
                "error": "Could not find a typescript root directory in this project.",
            }
        runner = ToolRunner(root)
        tsc_issues = runner.run(["npx", "tsc", "--noEmit"])
        eslint_issues = runner.run(["npx", "eslint", "--ext", ".js,.jsx,.ts,.tsx"])
        return {
            "status": "ok",
            "diagnostics": [
                {"source": "tsc", "output": tsc_issues},
                {"source": "eslint", "output": eslint_issues},
            ],
        }


class GoStrategy(LanguageStrategy):
    def __init__(self, project_root: str):
        super().__init__(project_root)

        self.available_tools = {"vet": self.run_go_vet}

    @property
    def configured_tools(self) -> List[Any]:
        return GoToolSettings().GO_TOOLS

    def run_go_vet(self):
        root = find_go_module_root(self.project_root)
        if not root:
            return Diagnostic(
                status="failure",
                source="vet",
                output="Could not find go.mod file for the project",
            )

        logger.debug(f"Analyzing entire Go project at root: {root}")
        runner = ToolRunner(root, GoToolSettings().BIN_DIR)
        vet_issues = runner.run(["go", "vet", "-json", "./..."])
        return Diagnostic(status="ok", source="vet", output=vet_issues)


class PythonStrategy(LanguageStrategy):
    def __init__(self, project_root: str):
        super().__init__(project_root)
        self.venv_path = find_virtual_env(self.project_root)
        if self.venv_path:
            logger.debug(f"Found Python virtual environment at: {self.venv_path}")
        else:
            logger.warning("No virtual environment found.")

        self.available_tools = {"ruff": self.run_ruff, "pyright": self.run_pyright}

    @property
    def configured_tools(self) -> List[Any]:
        return PythonToolSettings().TOOLS

    def run_ruff(self):
        ruff_executable = "ruff"

        if self.venv_path:
            ruff_venv_executable = os.path.join(self.venv_path, "bin", "ruff")
            if os.path.exists(ruff_venv_executable):
                ruff_executable = ruff_venv_executable

        runner = ToolRunner(self.project_root, PythonToolSettings().BIN_DIR)
        ruff_issues = runner.run(
            [ruff_executable, "check", ".", "--output-format=json", "--force-exclude"],
            parser=parse_as_json_document,
        )

        return Diagnostic(status="ok", source="ruff", output=ruff_issues)

    def run_pyright(self):
        pyright_cmd = ["npx", "pyright", "--outputjson"]

        if self.venv_path:
            pyright_cmd.extend(
                ["--pythonpath", os.path.join(self.venv_path, "bin", "python")]
            )

        runner = ToolRunner(self.project_root, PythonToolSettings().BIN_DIR)
        pyright_issues = runner.run(pyright_cmd, parser=parse_pyright_output)

        return Diagnostic(status="ok", source="pyright", output=pyright_issues)


LANGUAGE_STRATEGIES: Dict[str, Type[LanguageStrategy]] = {
    "go": GoStrategy,
    "python": PythonStrategy,
    "javascript": TypescriptStrategy,
    "typescript": TypescriptStrategy,
}



import atexit
import signal
import subprocess
import sys

from .server import mcp


def start_langtools_daemon():
    proc = subprocess.Popen(
        [sys.executable, "-u", "-m", "langtools_mcp.langtools_daemon.main"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    def cleanup():
        print("Shutting down langtools_daemon...")
        proc.terminate()
        try:
            proc.wait(timeout=5)
        except subprocess.TimeoutExpired:
            proc.kill()

    atexit.register(cleanup)

    def sig_handler(signum, frame):
        cleanup()
        sys.exit(0)

    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGINT, sig_handler)
    return proc

def main():
    print("Starting langtools_daemon sidecar...")
    _daemon_proc = start_langtools_daemon()
    mcp.run()


if __name__ == "__main__":
    main()



import logging
import threading

logger = logging.getLogger(__name__)


class LSPServerPool:
    def __init__(self, adapter_classes):
        """
        adapter_classes: dict mapping language string to LSP adapter class.
        Example: {"go": GoplsLSPAdapter, ...}
        """
        self.adapter_classes = adapter_classes
        self.lock = threading.Lock()
        self.servers = {}  # (language, root) -> adapter instance

    def get_server(self, language, root_path, **kwargs):
        with self.lock:
            key = (language, root_path)
            if key not in self.servers:
                logger.info(
                    f"Spawning new {language} LSP for root={root_path}",
                )
                adapter_cls = self.adapter_classes[language]
                self.servers[key] = adapter_cls(root_path=root_path, **kwargs)
            return self.servers[key]

    def shutdown(self):
        with self.lock:
            for srv in self.servers.values():
                srv.shutdown()
            self.servers.clear()
